{*------------------------------------------------------------------------------
  This unit should provide functions for easy
  generation and application of binary patches to program files.

  It is based on the program bsdiff from Colin Percival.
  Find the original source and license here:
  http://www.daemonology.net/bsdiff/

  The modified format (based on BSDIFF40) is:
    0 8 "ICTDIFF1"
    8 8 X
    16  8 Y
    24  8 sizeof(newfile)
    32 8 Date and Time of the new file (Ticks, Int64)
    40 32 Md5sum of old file
    72 8 FileVersion of the old dll
    80  X bzip2(control block)
    80+X  Y bzip2(diff block)
    80+X+Y  ??? bzip2(extra block)

  with control block a set of triples (x,y,z) meaning "add x bytes
  from oldfile to x bytes from the diff block; copy y bytes from the
  extra block; seek forwards in oldfile by z bytes".

  This Delphi implementation does only support BSDIFF40, version 2;
  and I have added new fields to the header as documented above, so it is not compatible anymore

  it also makes use of the SharpZiblib,
  http://www.icsharpcode.net/OpenSource/SharpZipLib/

  @Author  Timotheus Pokorra
  feel free to use this code; I guess the license from the original bsdiff applies

  @Version $Revision: 1.8 $ / $Date: 2007/06/27 12:49:01 $
-------------------------------------------------------------------------------}
unit bsdiff;

interface

uses
  System.IO,
  System.Diagnostics,
  System.Collections,
  ICSharpCode.SharpZipLib.BZip2;

const
  UNIT_REVISION = '$Revision: 1.8 $';

type TPatchTools = class

protected

private  
  FBsDiffFinished: Boolean;  
  // to be called after the c program for generating the patch has finished
  procedure BsDiffProcessEnd(ASender: TObject; AEventArgs: EventArgs);

  // required for creating the binary diff
  procedure Split(var AI, AV: array of Int64; start, len, h: Int32);
  // required for creating the binary diff
  procedure qsufsort(var AI, AV: array of Int64;old: array of byte; oldsize: Int32);
  // required for creating the binary diff
  function matchlen(old: array of byte; offsetold, oldsize: Int32; pnew: array of byte; offsetnew, newsize: Int32): Int32;
  // required for creating the binary diff
  function memcmp(old: array of byte; offsetold: Int32; new: array of byte; offsetnew, lengthOld, lengthNew: Int32): Int32;
  // required for creating the binary diff
  function search(AI: array of Int64; old: array of byte; oldsize: Int32;
    pnew: array of byte; offsetnew, newsize,st,en: Int64; var pos: Int32): Int32;
  // required for creating the binary diff
  procedure CalculateDiffAndCtrl(
      var zipStream: BZip2OutputStream;
      var old, pnew: array of byte;
      oldsize, newsize: Int32;
      var AI: array of Int64;
      var db, eb: array of byte;
      var dblen, eblen: Int32);

  procedure CreateDiff_External(AFileName1, AFileName2: String;
    var cb, db, eb: array of byte; var cblen, dblen, eblen: Int32);

  // required for creating the binary diff
  procedure offtout(x: Int64; var buf: array of byte);
  // required for applying a binary patch
  function offtin(buf: array of byte; offset: Int32): Int64;

  // required for applying a binary patch
  function loopread(var zipStream: BZip2InputStream; var buf: array of byte; offset, nbytes: Int32): Int32;
  // required for applying a binary patch
  function bz2read(var bz: BZip2InputStream; offset: Int32; fname: string): FileStream;

public
  constructor Create(); overload;

  // Create a binary diff between AFileName1 and AFileName2, and store the differences in AFileNameDiff
  // this version uses the C bsdiff program to generate the patch; much faster
  procedure CreateDiff(AFileName1, AFileName2, AFileNameDiff: String);
  // Create a binary diff between AFileName1 and AFileName2, and store the differences in AFileNameDiff
  // this is quite slow and memory consuming
  procedure CreateDiffManaged(AFileName1, AFileName2, AFileNameDiff: String);
  // Apply a patch to an existing file, and create a new file
  function ApplyPatch(AOldFile, ANewFile, APatchFile: String): Boolean;

  function CheckMd5Sum(ADLLName: String; AExpectedMd5Sum: String): Boolean;
  function GetMd5Sum(ADLLName: String): String;
  // compare the size and md5sum of files
  function IsSame(AFileName1, AFileName2: String): Boolean;

end;

implementation

uses
  System.Collections.Specialized,
  System.Threading,
  System.Runtime.InteropServices,
  System.Security.Cryptography;

const
  BZ_COMPRESSION_LEVEL = 5; // maximum is 9, quite slow, and memory consuming
  HEADER_SIZE = 80; // our own user defined header for patch files
  FORMAT_DESCR = 'ICTDIFF1';

constructor TPatchTools.Create();
begin
  inherited;
end;

procedure TPatchTools.Split(var AI, AV: array of Int64; start, len, h: Int32);
var
  i,j,k,x,tmp,jj,kk: Int32;
begin
  if(len<16) then
  begin
    k := start;
    while k < start+len do
    begin
      j:=1;
      x:=AV[AI[k]+h];
      i:=1;
      while (k+i<start+len) do
      begin
        if(AV[AI[k+i]+h]<x) then
        begin
          x:=AV[AI[k+i]+h];
          j:=0;
        end;
        if(AV[AI[k+i]+h]=x) then
        begin
          tmp:=AI[k+j];
          AI[k+j]:=AI[k+i];
          AI[k+i]:=tmp;
          Inc(j);
        end;
        Inc(i);
      end;
      for i:=0 to j-1 do
      begin
        AV[AI[k+i]]:=k+j-1;
      end;
      if(j=1) then
      begin
        AI[k]:=-1;
      end;
      k := k + j;
    end;
    exit;
  end;

  x:=AV[AI[start+len div 2]+h];
  jj:=0;kk:=0;
  i:=start;
  while (i<start+len) do
  begin
    if(AV[AI[i]+h]<x) then
    begin
      inc(jj);
    end;
    if(AV[AI[i]+h]=x) then
    begin
      inc(kk);
    end;
    Inc(i);
  end;
  jj := jj + start;
  kk := kk + jj;

  i:=start;j:=0;k:=0;
  while(i<jj) do
  begin
    if(AV[AI[i]+h]<x) then
    begin
      inc(i);
    end
    else if(AV[AI[i]+h]=x) then
    begin
      tmp:=AI[i];AI[i]:=AI[jj+j];AI[jj+j]:=tmp;
      Inc(j);
    end
    else
    begin
      tmp:=AI[i];AI[i]:=AI[kk+k];AI[kk+k]:=tmp;
      inc(k);
    end;
  end;

  while(jj+j<kk) do
  begin
    if(AV[AI[jj+j]+h]=x) then
    begin
      inc(j);
    end
    else
    begin
      tmp:=AI[jj+j];AI[jj+j]:=AI[kk+k];AI[kk+k]:=tmp;
      inc(k);
    end;
  end;

  if(jj>start) then
  begin
    split(AI,AV,start,jj-start,h);
  end;

  i:=0;
  while (i<kk-jj) do
  begin
    AV[AI[jj+i]]:=kk-1;
    inc(i);
  end;
  if(jj=kk-1) then
  begin
    AI[jj]:=-1;
  end;

  if(start+len>kk) then
  begin
    split(AI,AV,kk,start+len-kk,h);
  end;
end;

procedure TPatchTools.qsufsort(var AI, AV: array of Int64;old: array of byte; oldsize: Int32);
var
  buckets: array [0..256] of Int64;
  i,h,len: Int32;
begin
  for i:=0 to 255 do
  begin
    buckets[i]:=0;
  end;
  for i:=0 to oldsize-1 do
  begin
    inc(buckets[old[i]]);
  end;
  for i:=1 to 255 do
  begin
    buckets[i] := buckets[i] + buckets[i-1];
  end;
  for i:=255 downto 1 do
  begin
    buckets[i] := buckets[i-1];
  end;
  buckets[0]:=0;

  for i:=0 to oldsize - 1 do
  begin
    Inc(buckets[old[i]]);
    AI[buckets[old[i]]]:=i;
  end;
  AI[0]:=oldsize;
  for i:=0 to oldsize - 1 do
  begin
    AV[i]:=buckets[old[i]];
  end;
  AV[oldsize]:=0;
  for i:=1 to 255 do
  begin
    if(buckets[i]=buckets[i-1]+1) then
    begin
      AI[buckets[i]]:=-1;
    end;
  end;
  AI[0]:=-1;

  h:=1;
  while (AI[0]<>-(oldsize+1)) do
  begin
    len:=0;
    i:=0;
    while (i<oldsize+1) do
    begin
      if(AI[i]<0) then
      begin
        len:= len - AI[i];
        i := i - AI[i];
      end
      else
      begin
        if(len<>0) then
        begin
          AI[i-len]:=-len;
        end;
        len:=AV[AI[i]]+1-i;
        split(AI,AV,i,len,h);
        i := i + len;
        len:=0;
      end;
    end;
    if(len <> 0) then
    begin
      AI[i-len]:=-len;
    end;
    h := h+h;
  end;

  for i:=0 to oldsize do
  begin
    AI[AV[i]]:=i;
  end;
end;

function TPatchTools.matchlen(old: array of byte; offsetold, oldsize: Int32; pnew: array of byte; offsetnew, newsize: Int32): Int32;
var
  i: Int32;
begin
  i:=0;
  while (i<oldsize) and (i<newsize) do
  begin
    if(old[offsetold+i]<>pnew[offsetnew+i]) then
    begin
      result := i;
      exit;
    end;
    inc(i);
  end;
  result := i;
end;

function TPatchTools.memcmp(old: array of byte; offsetold: Int32; new: array of byte; offsetnew, lengthOld, lengthNew: Int32): Int32;
var
  i, length: Int32;
begin
  result := 0;
  if lengthOld < lengthNew then
  begin
    length := lengthOld;
  end
  else
  begin
    length := lengthNew;
  end;
  for i := 0 to length - 1 do
  begin
    if (old[offsetold+i] <> new[offsetnew+i]) then
    begin
      if (old[offsetold+i] > new[offsetnew+i]) then
      begin
        result := 1;
      end
      else
      begin
        result := -1;
      end;
      exit;
    end;
  end;
end;

function TPatchTools.search(AI: array of Int64; old: array of byte; oldsize: Int32;
    pnew: array of byte; offsetnew, newsize,st,en: Int64; var pos: Int32): Int32;
var
  x,y: Int64;
begin
  if(en-st<2) then
  begin
    x:=matchlen(old, AI[st],oldsize-AI[st],pnew,offsetnew,newsize);
    y:=matchlen(old, AI[en],oldsize-AI[en],pnew,offsetnew,newsize);
    if(x>y) then
    begin
      pos := AI[st];
      result := x;
      exit;
    end
    else
    begin
      pos :=AI[en];
      result := y;
      exit;
    end;
  end;

  x:=st+(en-st) div 2;

  if(memcmp(old,AI[x],pnew,offsetnew,oldsize-AI[x],newsize)<0) then
  begin
//Console.WriteLine('before search B '+ x.tostring+ ' ' +  en.tostring);
    result := search(AI,old,oldsize,pnew,offsetnew,newsize,x,en,pos);
    exit;
  end
  else
  begin
//Console.WriteLine('before search C '+ st.tostring+ ' '  +x.tostring)  ;
    result := search(AI,old,oldsize,pnew,offsetnew,newsize,st,x,pos);
    exit;
  end;
end;

procedure TPatchTools.offtout(x: Int64; var buf: array of byte);
var
  y: Int64;
begin
  y:=x;
  if(x<0) then
  begin
    y:=-x;
  end;
  
  buf[0]:=y mod 256;
  y := y - buf[0];
  y:=y div 256;buf[1]:=y mod 256;y := y - buf[1];
  y:=y div 256;buf[2]:=y mod 256;y := y - buf[2];
  y:=y div 256;buf[3]:=y mod 256;y := y - buf[3];
  y:=y div 256;buf[4]:=y mod 256;y := y - buf[4];
  y:=y div 256;buf[5]:=y mod 256;y := y - buf[5];
  y:=y div 256;buf[6]:=y mod 256;y := y - buf[6];
  y:=y div 256;buf[7]:=y mod 256;

  if(x<0) then
  begin
    buf[7] := buf[7] or $80;
  end;
end;

function TPatchTools.offtin(buf: array of byte; offset: Int32): Int64;
var
  y: Int64;
begin
  y:=buf[offset+7] and $7F;
  y:=y*256;y:=y+buf[offset+6];
  y:=y*256;y:=y+buf[offset+5];
  y:=y*256;y:=y+buf[offset+4];
  y:=y*256;y:=y+buf[offset+3];
  y:=y*256;y:=y+buf[offset+2];
  y:=y*256;y:=y+buf[offset+1];
  y:=y*256;y:=y+buf[offset+0];

  if((buf[offset+7] and $80) <> 0) then
  begin
    y:=-y;
  end;

  result := y;
end;

function TPatchTools.loopread(var zipStream: BZip2InputStream; var buf: array of byte; offset, nbytes: Int32): Int32;
var
  ptr, lenread: Int32;
begin
  ptr:=0;
  while (ptr<nbytes) do
  begin
    lenread:=zipStream.Read(buf,offset+ptr,nbytes);
    if(lenread=0) then
    begin
      result := ptr;
      exit;
    end;
    if(lenread=-1) then
    begin
      result := -1;
      exit;
    end;
    ptr:=ptr+lenread;
  end;
  result := ptr;
end;

function TPatchTools.bz2read(var bz: BZip2InputStream; offset: Int32; fname: string): FileStream;
var
  fs: FileStream;
begin
  try
    fs := FileStream.Create(fname, Filemode.Open,FileAccess.Read, FileShare.Read);
    fs.Seek(offset, SeekOrigin.Begin);
    bz := BZip2InputStream.Create(fs);
    result := fs;
  except on e: Exception do
    begin
      raise Exception.Create('Cannot open file ' + fname + ' for reading; '+e.Message);
    end;
  end;
end;

procedure TPatchTools.CalculateDiffAndCtrl(
  var zipStream: BZip2OutputStream;
  var old, pnew: array of byte;
  oldsize, newsize: Int32;
  var AI: array of Int64;
  var db, eb: array of byte;
  var dblen, eblen: Int32);
var
  buf: array[0..7] of byte;
  scan,pos,len: Int32;
  lastscan,lastpos,lastoffset: Int32;
  oldscore,scsc: Int64;
  s,Sf,lenf,Sb,lenb: Int64;
  overlap,Ss,lens: Int64;
  i: Int32;
  Progress: String;
begin
  scan:=0;len:=0;
  progress := '';
  lastscan:=0;lastpos:=0;lastoffset:=0;pos:=-1;
//  Console.Writeline('newsize: ' + newsize.tostring());
  while(scan<newsize) do
  begin
    oldscore:=0;

    scan := scan + len;
    scsc:=scan;
    while (scan<newsize) do
    begin
//Console.writeLine('before search '+scan.tostring() + ' ' + pos.tostring + ' ' + scsc.tostring);
      len:=search(AI,old,oldsize,pnew,scan,newsize-scan,
          0,oldsize,pos);
//Console.writeLine('after search ' +len.tostring + ' ' + pos.tostring);
      while(scsc<scan+len) do
      begin
        if((scsc+lastoffset<oldsize) and
          (old[scsc+lastoffset] = pnew[scsc])) then
        begin
          inc(oldscore);
        end;
        inc(scsc);
      end;

      if(((len=oldscore) and (len<>0)) or (len>oldscore+8)) then
      begin
        break;
      end;

      if((scan+lastoffset<oldsize) and (old[scan+lastoffset] = pnew[scan])) then
      begin
        dec(oldscore);
      end;
      inc(scan);
    end;

    if((len<>oldscore) or (scan=newsize)) then
    begin
      s:=0;Sf:=0;lenf:=0;
      i:=0;
      while (lastscan+i<scan) and (lastpos+i<oldsize) do
      begin
        if(old[lastpos+i]=pnew[lastscan+i]) then
        begin
          inc(s);
        end;
        inc(i);
        if(s*2-i>Sf*2-lenf) then
        begin
          Sf:=s; lenf:=i; 
        end;
      end;

      lenb:=0;
      if(scan<newsize) then
      begin
        s:=0;Sb:=0;
        i:=1;
        while(scan>=lastscan+i) and (pos>=i) do
        begin
          if(old[pos-i]=pnew[scan-i]) then
          begin
            inc(s);
          end;
          if(s*2-i>Sb*2-lenb) then
          begin
            Sb:=s; 
            lenb:=i;
          end;
          inc(i);
        end;
      end;

      if(lastscan+lenf>scan-lenb) then
      begin
        overlap:=(lastscan+lenf)-(scan-lenb);
        s:=0;Ss:=0;lens:=0;
        for i:=0 to overlap-1 do
        begin
          if(pnew[lastscan+lenf-overlap+i]=old[lastpos+lenf-overlap+i]) then
          begin
            inc(s);
          end;
          if(pnew[scan-lenb+i]=old[pos-lenb+i]) then
          begin
            dec(s);
          end;
          if(s>Ss) then
          begin
            Ss:=s; lens:=i+1;
          end;
        end;

        lenf := lenf+lens-overlap;
        lenb := lenb-lens;
      end;

      for i:=0 to lenf-1 do
      begin
        db[dblen+i]:=pnew[lastscan+i]-old[lastpos+i];
      end;
      for i:=0 to (scan-lenb)-(lastscan+lenf)-1 do
      begin
        eb[eblen+i]:=pnew[lastscan+lenf+i];
      end;

      dblen := dblen+lenf;
      eblen := eblen+(scan-lenb)-(lastscan+lenf);

      offtout(lenf,buf);
      zipStream.Write(buf,0,8);
      offtout((scan-lenb)-(lastscan+lenf),buf);
      zipStream.Write(buf, 0, 8);
      offtout((pos-lenb)-(lastpos+lenf),buf);
      zipStream.Write(buf, 0, 8);

      // to make the progress visible in the test program on the command line
      Progress := Progress + '.';
//      Console.Writeline(Progress + ' ' + Convert.ToString((scan*100) div newsize) + '%', [ToConsole]);

      lastscan:=scan-lenb;
      lastpos:=pos-lenb;
      lastoffset:=pos-scan;
    end;
  end;
end;

procedure TPatchTools.BsDiffProcessEnd(ASender: TObject; AEventArgs: EventArgs);
begin
  FBsDiffFinished := true;
end;

procedure TPatchTools.CreateDiff_External(AFileName1, AFileName2: String;
  var cb, db, eb: array of byte; var cblen, dblen, eblen: Int32);
var
  BSDIFFProcess: System.Diagnostics.Process;
  header: array[0..31] of byte;
  buf: array[0..7] of byte;
  r: BinaryReader;
  fsTemp: FileStream;
  i: Int32;
begin
  // make an external call to bsdiff, it will create a BSDIFF4.0 patch file
  BSDIFFProcess := System.Diagnostics.Process.Create();
  BSDIFFProcess.EnableRaisingEvents := false;
  BSDIFFProcess.StartInfo.FileName := 'bsdiff.exe'; // the program needs to be on the PATH
  BSDIFFProcess.StartInfo.Arguments := AFileName1 + ' ' + AFileName2 + ' temp.patch';
  BSDIFFProcess.StartInfo.WindowStyle := ProcessWindowStyle.Hidden;
  BSDIFFProcess.EnableRaisingEvents := true;

  FBsDiffFinished := false;
  if (BSDIFFProcess.Start()) then
  begin
    Include(BSDIFFProcess.Exited, BsDiffProcessEnd);
  end;

  while (not FBsDiffFinished) do
  begin
    Thread.Sleep(500);
  end;

  try 
    fsTemp := FileStream.Create('temp.patch', FileMode.Open, FileAccess.Read);
  except on e: Exception do
    begin
      raise Exception.Create('Cannot open the temporary file generated by bsdiff');
    end;
  end;

  r := BinaryReader.Create(fsTemp);
  // read the standard BSDIFF header
  header := r.ReadBytes(32);
  if (Length(header) <> 32) then
  begin
    raise Exception.Create('Corrupt temp patch');
  end;
  for i := 0 to 7 do
  begin
    if (header[i] <> Byte(String('BSDIFF40').Chars[i])) then
    begin
      raise Exception.Create('wrong version');
    end;
  end;

  cblen:=offtin(header,8);
  dblen:=offtin(header,16);
  eblen:=fsTemp.Length-(32+cblen+dblen);
  
  cb := r.ReadBytes(cblen);
  db := r.ReadBytes(dblen);
  eb := r.ReadBytes(eblen);

  r.Close();
  fsTemp.Close();
end;

procedure TPatchTools.CreateDiff(AFileName1, AFileName2, AFileNameDiff: String);
var
  fs1, fs2, fsDiff: FileStream;
  w: BinaryWriter;
  
  newsize: Int32;

  cb, db, eb: array of byte;
  cblen, dblen, eblen: Int32;
  header: array[0..HEADER_SIZE-1] of byte;
  buf: array[0..7] of byte;

  i: Int32;

  ver: FileVersionInfo;
  oldMd5Sum: String;
  
begin
  CreateDiff_External(AFileName1, AFileName2, cb, db, eb, cblen, dblen, eblen);

  try
    // open it, even though we don't need the file; just to be sure the external diff could access it as well
    fs1 := FileStream.Create(AFileName1, FileMode.Open, FileAccess.Read);
  except on e: Exception do
    begin
      raise Exception.Create('Cannot open file ' + AFileName1 + ' for reading');
    end;
  end;
  try
    fs2 := FileStream.Create(AFileName2, FileMode.Open, FileAccess.Read);
  except on e: Exception do
    begin
      raise Exception.Create('Cannot open file ' + AFileName2 + ' for reading');
    end;
  end;
  try
    fsDiff := FileStream.Create(AFileNameDiff, FileMode.Create);
  except on e: Exception do
    begin
      raise Exception.Create('Cannot write to file ' + AFileNameDiff);
    end;
  end;

  //oldsize := fs1.Length;
  fs1.Close;

  newsize := fs2.Length;
  fs2.Close;

  w := BinaryWriter.Create(fsDiff);
  // for header format see the unit description
  for i := 0 to 7 do
  begin
    header[i] := Byte(String(FORMAT_DESCR).Chars[i]);
  end;
  for i := 8 to HEADER_SIZE-1 do
  begin
    header[i] := 0;
  end;
  w.write(header);

  // insert the control block
  w.Write(cb, 0, cblen);
  // write the diff block
  w.Write(db, 0, dblen);
  // write the extra block
  w.Write(eb, 0, eblen);

  // write the lengths of each block to the header
  w.Seek(8,SeekOrigin.Begin);
  offtout(cblen,buf);
  w.Write(buf, 0, 8);
  w.Seek(16,SeekOrigin.Begin);
  offtout(dblen,buf);
  w.Write(buf, 0, 8);
  w.Seek(24,SeekOrigin.Begin);
  offtout(newsize,buf);
  w.Write(buf, 0, 8);

  // write the last modification time of the new file
  offtout(System.IO.File.GetLastWriteTime(AFileName2).Ticks, buf);
  w.Write(buf, 0, 8);

  // write the md5sum of the old file
  oldMd5Sum := GetMd5Sum(AFileName1);
  for i := 0 to 31 do
  begin
    w.Write(Convert.ToByte(oldMd5Sum.Chars[i]));
  end;

  // file version of the old dll
  ver := FileVersionInfo.GetVersionInfo(AFileName1);
  // each part of the version is a 16 bit value; together the version is saved in 64 bites
  w.Write(Convert.ToByte(ver.FileMajorPart div 256));
  w.Write(Convert.ToByte(ver.FileMajorPart mod 256));
  w.Write(Convert.ToByte(ver.FileMinorPart div 256));
  w.Write(Convert.ToByte(ver.FileMinorPart mod 256));
  w.Write(Convert.ToByte(ver.FileBuildPart div 256));
  w.Write(Convert.ToByte(ver.FileBuildPart mod 256));
  w.Write(Convert.ToByte(ver.FilePrivatePart div 256));
  w.Write(Convert.ToByte(ver.FilePrivatePart mod 256));

  w.Close();
  fsDiff.Close();
end;

procedure TPatchTools.CreateDiffManaged(AFileName1, AFileName2, AFileNameDiff: String);
var
  fs1, fs2, fsDiff: FileStream;
  oldsize, newsize: Int32;
  old, pnew: array of byte;
  r: BinaryReader;
  w: BinaryWriter;
  AI, AV: array of int64;
  db, eb: array of byte;
  dblen, eblen: Int32;
  header: array[0..HEADER_SIZE-1] of byte;
  buf: array[0..7] of byte;

  zipStream: BZip2OutputStream;

  lengthCtrl, lengthDiff, i: Int32;

  ver: FileVersionInfo;
  oldMd5Sum: String;
begin
  try
    fs1 := FileStream.Create(AFileName1, FileMode.Open, FileAccess.Read);
  except on e: Exception do
    begin
      raise Exception.Create('Cannot open file ' + AFileName1 + ' for reading');
    end;
  end;
  try
    fs2 := FileStream.Create(AFileName2, FileMode.Open, FileAccess.Read);
  except on e: Exception do
    begin
      raise Exception.Create('Cannot open file ' + AFileName2 + ' for reading');
    end;
  end;
  try
    fsDiff := FileStream.Create(AFileNameDiff, FileMode.Create);
  except on e: Exception do
    begin
      raise Exception.Create('Cannot write to file ' + AFileNameDiff);
    end;
  end;

  oldsize := fs1.Length;
  r := BinaryReader.Create(fs1);
  old := r.ReadBytes(oldSize);
  r.Close;
  fs1.Close;

  newsize := fs2.Length;
  r := BinaryReader.Create(fs2);
  pnew := r.ReadBytes(newSize);
  {$IFDEF DEBUGMODE}
  Console.Writeline('after loading both files');
  {$ENDIF}
  r.Close;
  fs2.Close;

  try
    // at the moment, the creating of the diff is too slow; would it help to work with pointers and unmanaged or pinned memory?
    //AIPtr := Marshal.AllocHGlobal(oldsize+1); IntPtr
    //Marshal.FreeHGlobal
    //AVPtr := Marshal.AllocHGlobal(oldsize+1);
    //http://www.sdss.jhu.edu/~tamas/bytes/fftwcsharp.html hin hout AddrOfPinnedObject
    SetLength(AI, oldsize+1);
    SetLength(AV, oldsize+1);
    SetLength(db, newsize+1);
    SetLength(eb, newsize+1);
  except
    on e: Exception do
    begin
      Console.Writeline('problem reserving memory ' + Convert.Tostring(oldsize*2 + newsize*2));
      raise e;
    end;
  end;
  {$IFDEF DEBUGMODE}
  Console.Writeline('after reserving memory');
  {$ENDIF}

  for i := 0 to oldsize do
  begin
    AI[i] := 0;
    AV[i] := 0;
  end;
  qsufsort(AI,AV,old,oldsize);
  SetLength(AV, 0);
  {$IFDEF DEBUGMODE}
  Console.Writeline('after qsufsort');
  {$ENDIF}

  dblen:=0;
  eblen:=0;

  w := BinaryWriter.Create(fsDiff);

  // for header format see the unit description
  for i := 0 to 7 do
  begin
    header[i] := Byte(String(FORMAT_DESCR).Chars[i]);
  end;
  for i := 8 to HEADER_SIZE-1 do
  begin
    header[i] := 0;
  end;
  w.write(header);

  // calculate the diff and extra block, write the control block
  zipStream := BZip2OutputStream.Create(fsDiff, BZ_COMPRESSION_LEVEL);
  zipStream.IsStreamOwner := false;
  CalculateDiffAndCtrl(zipStream, old, pnew, oldsize, newsize, AI, db, eb, dblen, eblen);
  zipStream.Close();
  lengthCtrl := zipstream.BytesWritten;

  // write the diff block
  zipStream := BZip2OutputStream.Create(fsDiff, BZ_COMPRESSION_LEVEL);
  zipStream.IsStreamOwner := false;
  zipStream.Write(db, 0, dblen);
  zipStream.Close();
  lengthDiff := zipstream.BytesWritten;

  // write the control block
  zipStream := BZip2OutputStream.Create(fsDiff, BZ_COMPRESSION_LEVEL);
  zipStream.IsStreamOwner := false;
  zipStream.Write(eb, 0, eblen);
  zipStream.Close();

  // write the lengths of each block to the header
  w.Seek(8,SeekOrigin.Begin);
  offtout(lengthCtrl,buf);
  w.Write(buf, 0, 8);
  w.Seek(16,SeekOrigin.Begin);
  offtout(lengthDiff,buf);
  w.Write(buf, 0, 8);
  w.Seek(24,SeekOrigin.Begin);
  offtout(newsize,buf);
  w.Write(buf, 0, 8);

  // write the last modification time of the new file
  offtout(System.IO.File.GetLastWriteTime(AFileName2).Ticks, buf);
  w.Write(buf, 0, 8);

  // write the md5sum of the old file
  oldMd5Sum := GetMd5Sum(AFileName1);
  for i := 0 to 31 do
  begin
    w.Write(Convert.ToByte(oldMd5Sum.Chars[i]));
  end;

  // file version of the old dll
  ver := FileVersionInfo.GetVersionInfo(AFileName1);
  // each part of the version is a 16 bit value; together the version is saved in 64 bites
  w.Write(Convert.ToByte(ver.FileMajorPart div 256));
  w.Write(Convert.ToByte(ver.FileMajorPart mod 256));
  w.Write(Convert.ToByte(ver.FileMinorPart div 256));
  w.Write(Convert.ToByte(ver.FileMinorPart mod 256));
  w.Write(Convert.ToByte(ver.FileBuildPart div 256));
  w.Write(Convert.ToByte(ver.FileBuildPart mod 256));
  w.Write(Convert.ToByte(ver.FilePrivatePart div 256));
  w.Write(Convert.ToByte(ver.FilePrivatePart mod 256));

  w.Close();
  fsDiff.Close();
end;

function TPatchTools.ApplyPatch(AOldFile, ANewFile, APatchFile: String): Boolean;
var
  bzCtrl, bzDiff, bzExtra: BZip2InputStream;
  fsCtrl, fsDiff, fsExtra: FileStream;
  fsHeader, fsOld, fsNew: FileStream;
  brHeader, brOld: BinaryReader;
  bwNew: BinaryWriter;
  patchsize,oldsize,newsize: Int32;
  bzctrllen,bzdatalen: Int32;

  header: array[0..HEADER_SIZE-1] of byte;
  buf: array[0..8] of byte;

  NewFileDateTime: DateTime;
  OldMd5sum: String;
  ver: FileVersionInfo;

  old, pnew: array of byte;

  i, version, oldpos, newpos, lenread: Int32;
  ctrl: array[0..2] of int32;
  StoredMajorPart, StoredMinorPart, StoredBuildPart, StoredPrivatePart: Int16;
begin
  fsHeader := FileStream.Create(APatchFile, FileMode.Open);
  brHeader := BinaryReader.Create(fsHeader);
  patchsize := fsHeader.length;
  if (patchsize < HEADER_SIZE) then
  begin
    raise Exception.Create('Corrupt patch (1)');
  end;

  // see header format in unit description
  header := brHeader.ReadBytes(HEADER_SIZE);
  brHeader.Close();
  fsHeader.Close();
  if (Length(header) <> HEADER_SIZE) then
  begin
    raise Exception.Create('Corrupt patch (2)');
  end;
  for i := 0 to 7 do
  begin
    if (header[i] <> Byte(String(FORMAT_DESCR).Chars[i])) then
    begin
      raise Exception.Create('wrong version');
    end;
  end;
  version := 2;

  bzctrllen:=offtin(header,8);
  bzdatalen:=offtin(header,16);
  newsize:=offtin(header,24);
  if((bzctrllen<0) or (bzdatalen<0) or (newsize<0)) then
  begin
    raise Exception.Create('Corrupt patch (3)');
  end;

  NewFileDateTime := DateTime.Create(offtin(header,32));
  OldMd5sum := '';
  for i := 0 to 31 do
  begin
    OldMd5sum := OldMd5Sum + Convert.ToChar(header[40+i]);    
  end;
  if (not CheckMd5Sum(AOldFile, OldMd5sum)) then
  begin
    raise Exception.Create('different base file, md5sum does not match');
  end;

  ver := FileVersionInfo.GetVersionInfo(AOldFile);
  StoredMajorPart := Int16(header[72])*256+header[73];
  StoredMinorPart := header[74]*256+header[75];
  StoredBuildPart := header[76]*256+header[77];
  StoredPrivatePart := header[78]*256+header[79];
  
  if (ver.FileMajorPart <> StoredMajorPart)
    or (ver.FileMinorPart <> StoredMinorPart)
    or (ver.FileBuildPart <> StoredBuildPart)
    or (ver.FilePrivatePart <> StoredPrivatePart) then
  begin
    raise Exception.Create('the existing file has an unexpected version number, expected '
      +StoredMajorPart.ToString() + '.'
      +StoredMinorPart.ToString() + '.'
      +StoredBuildPart.ToString() + '.'
      +StoredPrivatePart.ToString() + ', but was ' + ver.FileVersion);
  end;

  fsCtrl:=bz2read(&bzCtrl, HEADER_SIZE, APatchFile);
  fsDiff:=bz2read(&bzDiff, HEADER_SIZE+bzctrllen, APatchFile);
  fsExtra:=bz2read(&bzExtra, HEADER_SIZE+bzctrllen+bzdatalen, APatchFile);

  try
    fsOld := FileStream.Create(aOldFile, FileMode.Open);
    brOld := BinaryReader.Create(fsOld);
  except
    on e: Exception do
    begin
      raise Exception.Create('Cannot read file ' + aOldFile); 
    end;
  end;

  oldSize := fsOld.Length;
  old := brOld.ReadBytes(oldSize);
  if (length(old) <> oldSize) then
  begin
    raise Exception.Create('old file ' + aOldFile + ' has invalid size');
  end;
  brOld.Close;
  fsOld.Close;

  SetLength(pNew, newsize+1);

  oldpos:=0;newpos:=0;
  while(newpos<newsize) do
  begin
    for i:=0 to version do // we only support version 2
    begin
      lenread:=loopread(bzCtrl,buf,0,8);
      if (lenread<0) or (lenread<8) then
      begin
        raise Exception.Create('Corrupt patch (4)');
      end;
      ctrl[i]:=offtin(buf, 0);
    end;

    if(newpos+ctrl[0]>newsize) then
    begin
      raise Exception.Create('Corrupt patch (5)');
    end;
    lenread:=loopread(bzDiff,pnew,newpos,ctrl[0]);

    if (lenread<0) or (lenread<>ctrl[0]) then
    begin
      raise Exception.Create('Corrupt patch (6)');
    end;
    for i:=0 to ctrl[0]-1 do
    begin
      if((oldpos+i>=0) and (oldpos+i<oldsize)) then
      begin
        pnew[newpos+i]:=pnew[newpos+i]+old[oldpos+i];
      end;
    end;
    newpos := newpos+ctrl[0];
    oldpos := oldpos+ctrl[0];

    if(version=2) then
    begin
      if(newpos+ctrl[1]>newsize) then
      begin
        raise Exception.Create('Corrupt patch (7)');
      end;
      lenread:=loopread(bzExtra,pnew,newpos,ctrl[1]);
      if (lenread<0) or (lenread<>ctrl[1]) then
      begin
        raise Exception.Create('Corrupt patch (8)');
      end;

      newpos:=newpos+ctrl[1];
      oldpos:=oldpos+ctrl[2];
    end;
  end;

  // make sure there is nothing left to read
  if(loopread(bzCtrl,buf,0,1)<>0) or (loopread(bzDiff,buf,0,1)<>0) or (loopread(bzExtra,buf,0,1)<>0) then
  begin
    raise Exception.Create('Corrupt patch (9)');
  end;

  bzCtrl.Close();
  bzDiff.Close();
  bzExtra.Close();
  fsCtrl.Close();
  fsDiff.Close();
  fsExtra.Close();

  fsNew := FileStream.Create(ANewFile, FileMode.Create);
  bwNew := BinaryWriter.Create(fsNew);
  bwNew.Write(pnew, 0, newsize);
  bwNew.Close();
  fsNew.Close();
  System.IO.File.SetLastWriteTime(ANewFile, NewFileDateTime);

  result := true;
end;

function TPatchTools.GetMd5Sum(ADLLName: String): String;
var
  cr: MD5CryptoServiceProvider;
  fs: FileStream;
begin
  fs := FileStream.Create(ADLLName, FileMode.Open, FileAccess.Read);
  cr := MD5CryptoServiceProvider.Create();
  result := BitConverter.ToString(cr.ComputeHash(fs)).Replace('-','').ToLower();
  fs.Close();
end;

function TPatchTools.CheckMd5Sum(ADLLName: String; AExpectedMd5Sum: String): Boolean;
begin
  result := GetMd5Sum(ADLLName).Equals(AExpectedMd5Sum);
end;

function TPatchTools.IsSame(AFileName1, AFileName2: String): Boolean;
var
  size1, size2: Int64;
  fs: FileStream;
begin
  result := false;
  fs := System.IO.File.OpenRead(AFilename1);
  size1 := fs.Length;
  fs.Close();
  fs := System.IO.File.OpenRead(AFilename2);
  size2 := fs.Length;
  fs.Close();
  // only if size is the same make the effort to compare the checksums
  if (size1 = size2) then
  begin
    if (GetMd5Sum(AFileName1).Equals(GetMd5Sum(AFilename2))) then
    begin
      result := true;
    end;
  end;
end;

end.
